from orjson import loads
from typing import Optional, Type

from .exceptions import UpdateMalformed
from .model import BaseModel
from .transport import Transport
from .update import Update


class JSONTransport(Transport):
    type: Type = str

    ##################
    # Server methods #
    ##################
    async def onConnect(self, client_id: Optional[str] = None) -> str:
        # defer to parent
        return await super().onConnect(client_id=client_id)

    async def host(self, model: BaseModel, client_id: str, shared: bool = True, readonly: bool = False) -> None:
        # defer to parent
        return await super().host(model=model, client_id=client_id, shared=shared, readonly=readonly)

    async def onDisconnect(self, model: BaseModel, client_id: str) -> None:  # type: ignore[override]
        # defer to parent
        return await super().onDisconnect(model=model, client_id=client_id)

    def initial(self, client_id: str) -> str:  # type: ignore[override]
        # defer to parent
        return super().initial(client_id=client_id).json()

    ##################
    # Client methods #
    ##################
    async def onInitial(self, update: str) -> BaseModel:  # type: ignore[override]
        update_inst: Update = await self._update_to_model(update)
        return await super().onInitial(update=update_inst)

    async def _update_to_model(self, update: str) -> Update:
        # first, parse out the json into a dict
        data = loads(update)

        # now lookup the model map
        if "model_type" not in data:
            raise UpdateMalformed("Update data has no `model_type`")

        if "model_target" not in data:
            raise UpdateMalformed("Update data has no `model_target`")

        if data["model_type"] not in self.model_map:
            raise UpdateMalformed(f"Class type ({data['model_type']}) not known, did you forget to call `hosts`?")

        # grab the class type
        class_type: Type[BaseModel] = self.model_map[data.pop("model_type")]

        # instantiate the inital model
        model_inst: BaseModel = class_type.parse_obj(data["model"])
        data["model"] = model_inst

        # defer to parent now that we have the model
        return Update(**data)

    #################
    # Bidirectional #
    #################
    async def send(self, client_id: str) -> str:  # type: ignore[override]
        return (await super().send(client_id=client_id)).json()

    async def receive(self, client_id: str, update: str) -> None:  # type: ignore[override]
        update_inst: Update = await self._update_to_model(update)
        await super().receive(client_id=client_id, update=update_inst)



from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Optional, Type
from .transport import Transport
from .update import Update


class Connection(ABC):
    _transport: Transport
    _transport_type: Type
    _client_id: str

    @contextmanager
    def handleDisconnect(self):
        """Should be overloaded like so by specific client implementations:

        @contextmanager
        def handleDisconnect(self):
            with super().handleDisconnect():
                try:
                    yield
                except (ServerFrameworkSpecificExceptions...):
                    ...
        """
        try:
            yield
        except (KeyboardInterrupt,):
            ...

    async def receiver(self) -> None:
        with self.handleDisconnect():
            while True:
                # grab from client
                update: Update = await self.receive()

                # send to transport to be sent to models
                await self._transport.receive(client_id=self._client_id, update=update)

    async def sender(self) -> None:
        """An infinite async generator that should"""
        with self.handleDisconnect():
            while True:
                # get update from transport derived from models
                update: Update = await self._transport.send(client_id=self._client_id)

                # send to client
                await self.send(update)

    @abstractmethod
    async def connect(self, client_id: Optional[str] = None) -> str:
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        ...

    @abstractmethod
    async def receive(self) -> Update:
        ...

    @abstractmethod
    async def send(self, update: Update) -> None:
        ...



import time
import pytest
from typing import List, Dict
from transports import Transport, BaseModel


class MyModel(BaseModel):
    ...


class MyOtherModel(BaseModel):
    ...


class MyParentModel(BaseModel):
    x: List[MyModel]
    y: Dict[str, MyOtherModel]


class TestTransport:
    def test_instantiation(self):
        x = MyModel()
        assert x.name == x.id

    def test_copy(self):
        x = MyModel()

        # introduce a minimal delay
        time.sleep(0.001)

        y = x.copy()
        z = x.copy(clone=True)

        # copy assigns new id and new last_updated
        assert x.id != y.id
        assert x.created == y.created
        assert x.modified != y.modified

        # copy with cloning creates identical copy
        assert x.id == z.id
        assert x.created == z.created
        assert x.modified == z.modified

    def test_copy_freeze(self):
        x = MyModel()

        # introduce a minimal delay
        time.sleep(0.001)

        y = x.copy(freeze=True)

        # copy assigns new id and new last_updated
        assert x.id != y.id
        assert x.created == y.created
        assert x.modified != y.modified

        with pytest.raises(TypeError):
            y.name = "test"
        x.name = "test"
        assert y.name != "test"
        assert x.name == "test"

    def test_attach_transport(self):
        mm = MyModel()
        mom = MyOtherModel()
        pm = MyParentModel(x=[mm], y={"a": mom})

        transport = Transport()
        pm.onTransport(transport)

        assert pm._transport == transport
        assert mm._transport == transport
        assert mom._transport == transport

    def test_modelmap(self):
        ts = Transport(None)
        ts.hosts(MyParentModel)

        assert "MyParentModel" in ts.model_map
        assert ts.model_map["MyParentModel"] == MyParentModel
        assert "MyModel" in ts.model_map
        assert ts.model_map["MyModel"] == MyModel
        assert "MyOtherModel" in ts.model_map
        assert ts.model_map["MyOtherModel"] == MyOtherModel



from abc import ABCMeta
from asyncio import get_event_loop, run_coroutine_threadsafe, AbstractEventLoop
from typing import Dict, Optional, Type
from uuid import uuid4
from .model import BaseModel
from .update import Update


class ThreadedMixin:
    loop: AbstractEventLoop

    def run(self, coro):
        """Run a coroutine in a threadsafe and syncronous manner"""
        future = run_coroutine_threadsafe(coro, self.loop)
        # return raw future, don't wait for result
        return future.result()

    async def runAsync(self, coro):
        """Run a coroutine in a threadsafe and syncronous manner"""
        future = run_coroutine_threadsafe(coro, self.loop)
        # return raw future, don't wait for result
        return future


class Transport(ThreadedMixin, metaclass=ABCMeta):
    # Server attributes
    readonly: Dict[str, bool]  # Client -> T/F
    models: Dict[str, BaseModel]  # Client -> Model
    clients: Dict[str, str]  # Model ID -> Client
    type: Type = BaseModel

    # Client attributes
    server_models: Dict[str, BaseModel]  # type: ignore[no-redef]
    # Model ID -> Model
    # NOTE this is reused for both client and server
    model_map: Dict[str, Type[BaseModel]]  # Class name -> BaseModel type

    # General attributes
    loop: AbstractEventLoop

    def __init__(self, event_loop=None):
        # Server attributes
        self.readonly = {}
        self.models = {}
        self.clients = {}

        # Client attributes
        self.server_models = {}
        self.model_map = {}

        # General attributes
        self.loop = event_loop or get_event_loop()

    ##################
    # Server methods #
    ##################
    async def onConnect(self, client_id: Optional[str] = None) -> str:
        """Connect to a client.
        Should be called by the server when the client connects to the server.

        Args:
            client_id (Optional[str], optional): A unique string to represent the client, should be generated if not provided. Defaults to None.
        Returns:
            str: client id, either the provided value or the newly generated client id
        """
        if not client_id:
            client_id = str(uuid4())

        # and return the (possibly generated) client id
        return client_id

    async def host(self, model: BaseModel, client_id: str, shared: bool = True, readonly: bool = False) -> None:
        """Host a model for a given client.

        Args:
            model (BaseModel): An instance of BaseModel. Defaults to None.
            client_id (Optional[str], optional): A unique string to represent the client, should be generated if not provided. Defaults to None.
            shared (bool, optional): Whether or not the model should be shared amonst multiple clients.
                                     If False, the model will be copied. Defaults to True.
            readonly (bool, optional): Whether or not the client should be able to modify the model. Defaults to False.

        Returns:
            str: client id, either the provided value or the newly generated client id
        """
        if not shared or readonly:
            # copy the model
            model = model.copy()

        if readonly:
            # mark model as readonly
            model.freeze()

        # maintain map of client id -> model id
        self.models[client_id] = model

        # maintain reverse map of model id -> client id
        self.clients[model.id] = client_id

        # maintain map of if client's view is readonly
        self.readonly[client_id] = readonly

        # notify model of connection
        model.notifyConnect(client_id=client_id)

    async def onDisconnect(self, model: BaseModel, client_id: str) -> None:
        """Disconnect and cleanup client-specific assets.

        Should be called by the server when the client disconnects from the server.

        Args:
            client_id (str): _description_
        """
        model = self.models.pop(client_id, None) or model  # type: ignore[arg-type]
        self.readonly.pop(client_id, None)
        self.clients.pop(model.id, None)

        # notify model of disconnect
        model.notifyDisconnect(client_id=client_id)

    def initial(self, client_id: str) -> Update:
        """Create an initial update to be sent to the client

        Args:
            client_id (str): client id for looking up model to send

        Returns:
            Update: the initial model to send to the client
        """
        return Update(model=self.models[client_id])

    ##################
    # Client methods #
    ##################
    def hosts(self, model_type: Type[BaseModel]) -> None:
        """Host a model type, allowing the server to create instances of that type

        Args:
            model_type (Type[BaseModel]): type of model to create, should be unique
        """

        # register model as eligible
        self.model_map[model_type.__name__] = model_type

        # recursively descent types and register any other models we find
        for value_type in model_type._walk_types():
            if isinstance(value_type, type) and issubclass(value_type, BaseModel) and value_type.__name__ not in self.model_map:
                self.hosts(value_type)

    async def connect(self):
        # TODO
        ...

    async def disconnect(self):
        # TODO
        self.server_models = {}

    async def onInitial(self, update: Update) -> BaseModel:
        # TODO multi model and events
        model = update.model

        # register in model map
        self.server_models[model.id] = model

        # FIXME
        self.models[""] = model

        # return the instance
        return model

    #################
    # Bidirectional #
    #################
    async def send(self, client_id: str) -> Update:
        # grab model from client
        model = self.models[client_id]

        # pull latest send update
        return await model.getAsync()

    async def receive(self, client_id: str, update: Update) -> None:
        # grab model from client
        model = self.models[client_id]

        # push update to model
        await model.receiveAsync(update)



from aiohttp import ClientSession, ClientWebSocketResponse
from typing import Optional, Type
from ..client import Client
from ..json import JSONTransport


class AioHttpWebSocketClient(Client):
    _transport: JSONTransport
    _transport_type: Type = str

    _url: str
    _session: ClientSession
    _websocket: ClientWebSocketResponse
    _client_header: str

    def __init__(
        self,
        url: str,
        transport: JSONTransport,
        client_id: Optional[str] = None,
        client_header: str = "client-id",
        **kwargs,
    ):
        super().__init__(transport=transport, client_id=client_id)
        self._url = url
        self._client_header = client_header

    async def connect(self, client_id: Optional[str] = None):
        self._session = ClientSession()
        self._websocket = await self._session.ws_connect(
            self._url, headers={self._client_header: self._client_id} if self._client_id else None
        ).__aenter__()

    async def disconnect(self) -> None:
        await self._websocket.close()

    async def receive(self) -> str:  # type: ignore[override]
        return await self._websocket.receive_str()

    async def send(self, update: str) -> None:  # type: ignore[override]
        await self._websocket.send_str(update)



from .model import (
    BaseModel,
    # ListModel,
    # DictModel,
    Field,
    PrivateAttr,
)
from .transport import Transport
from .update import Update

from .json import JSONTransport
from .exceptions import UpdateMalformed

from .handlers import (
    # handlers
    StarletteWebSocketServer,
    # clients
    AioHttpWebSocketClient,
)


__version__ = "0.1.1"


__all__ = [
    "__version__",
    "BaseModel",
    "Field",
    "PrivateAttr",
    "Transport",
    "Update",
    "JSONTransport",
    "StarletteWebSocketServer",
    "AioHttpWebSocketClient",
]



import pydantic
from datetime import datetime
from udatetime import utcnow
from pydantic import BaseModel as PydanticBaseModel, Field, PrivateAttr  # noqa: F401
from typing import Optional
from uuid import uuid4
from .model import BaseModel
from .utils import orjson_dumps, orjson_loads


class Update(PydanticBaseModel):
    # Basic fields
    id: str = ""
    created: Optional[datetime] = None
    modified: Optional[datetime] = None

    # Advanced fields
    model: BaseModel

    # internals
    model_type: str = ""  # class type of `model`
    model_target: str = ""  # model target to which `model` should be sent to

    @pydantic.validator("id", pre=True, always=True)
    def default_id(cls, v, *, values, **kwargs):
        return v or str(uuid4())

    @pydantic.validator("created", pre=True, always=True)
    def default_created(cls, v):
        return v or utcnow()

    @pydantic.validator("modified", pre=True, always=True)
    def default_modified(cls, v, *, values, **kwargs):
        return v or values["created"]

    @pydantic.validator("model_target", pre=True, always=True)
    def default_model_target(cls, v, *, values, **kwargs):
        return v or values["model"].id

    # pydantic configuration
    class Config:
        # any types
        arbitrary_types_allowed = True

        # allow mutation (see frozen as well)
        allow_mutation = False

        # pass around models by value
        copy_on_model_validation = "deep"

        # no undeclared fields
        extra = "forbid"

        # frozen
        frozen = True

        # access by attr
        orm_mode = True

        # json_encoders = {}
        json_dumps = orjson_dumps
        json_loads = orjson_loads

    def __hash__(self):
        return hash(self.id)

    # overload dict to include model name
    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)
        ret["model_type"] = self.model.__class__.__name__
        return ret

    # `json()` doesnt use `dict`, so have to get creative
    def json(self, *args, **kwargs):
        return self.__config__.json_dumps(self.dict(), default=self.__json_encoder__)



from asyncio import gather
from typing import Optional, Type
from .connection import Connection
from .model import BaseModel
from .transport import Transport


class Client(Connection):
    _transport: Transport
    _transport_type: Type
    _model: BaseModel
    _client_id: str = ""

    def __init__(
        self,
        transport: Transport,
        client_id: Optional[str] = None,
        **kwargs,
    ):
        self._transport = transport
        self._transport_type = transport.type
        self._client_id = client_id or ""

    async def open(self) -> BaseModel:
        # wait for client connection
        await self.connect(client_id=self._client_id)

        # Notify transport of connection
        await self._transport.connect()

        # receive initial
        initial = await self.receive()

        # register and return
        return await self._transport.onInitial(initial)

    async def close(self):
        # Notify transport of connection
        await self._transport.disconnect()
        await self.disconnect()

    async def initial(self) -> BaseModel:
        # run the open routing
        return await self.open()

    async def handle(self) -> None:
        try:
            # Now enter an infinite loop listening for updates from client/server forever
            await gather(self.sender(), self.receiver(), return_exceptions=True)
        except:
            raise
        finally:
            # run the on close routing
            await self.close()



from contextlib import contextmanager
from starlette.websockets import WebSocket, WebSocketDisconnect
from typing import Type
from ..server import Server
from ..model import BaseModel
from ..json import JSONTransport


class StarletteWebSocketServer(Server):
    _transport: JSONTransport
    _transport_type: Type = str

    _websocket: WebSocket
    _client_header: str

    def __init__(
        self,
        websocket: WebSocket,
        transport: JSONTransport,
        model: BaseModel,
        shared: bool = True,
        readonly: bool = False,
        client_header: str = "client-id",
        **kwargs,
    ):
        super().__init__(transport=transport, model=model, shared=shared, readonly=readonly, **kwargs)
        self._websocket = websocket
        self._client_header = client_header

    async def connect(self):
        # on open, receive data from websocket
        # connect to websocket
        await self._websocket.accept()

        # graph client information from websocket,
        # if null or empty will be autoassigned
        return self._websocket.headers.get(self._client_header, "")

    @contextmanager
    def handleDisconnect(self):
        with super().handleDisconnect():
            try:
                yield
            except (WebSocketDisconnect,):
                ...

    async def receive(self) -> str:  # type: ignore[override]
        # grab from client
        return await self._websocket.receive_text()

    async def send(self, update: str):  # type: ignore[override]
        # send to client
        await self._websocket.send_text(update)

    async def disconnect(self) -> None:
        try:
            await self._websocket.close()
        except RuntimeError:
            # ignore if websocket already closed
            ...

    # Use common docstring
    __init__.__doc__ = Server.__init__.__doc__



import pydantic
from asyncio import Queue
from collections import deque
from datetime import datetime
from udatetime import utcnow
from pydantic import BaseModel as PydanticBaseModel, Field, PrivateAttr  # noqa: F401
from typing import Any, Optional, TYPE_CHECKING
from uuid import uuid4
from .utils import orjson_dumps, orjson_loads


if TYPE_CHECKING:
    from .transport import Transport
    from .update import Update


class BaseModel(PydanticBaseModel):
    # Basic fields
    id: str = ""
    name: str = ""
    label: str = ""

    # date fields
    created: Optional[datetime] = None
    modified: Optional[datetime] = None

    # internal fields
    _frozen: bool = PrivateAttr(False)

    # TODO make threadsafe
    _out_queue: Queue["Update"] = PrivateAttr(default_factory=Queue)

    @pydantic.validator("id", pre=True, always=True)
    def default_id(cls, v, *, values, **kwargs):
        return v or str(uuid4())

    @pydantic.validator("name", pre=True, always=True)
    def default_name(cls, v, *, values, **kwargs):
        return v or values["id"]

    @pydantic.validator("created", pre=True, always=True)
    def default_created(cls, v):
        return v or utcnow()

    @pydantic.validator("modified", pre=True, always=True)
    def default_modified(cls, v, *, values, **kwargs):
        return v or values["created"]

    # pydantic configuration
    class Config:
        # any types
        arbitrary_types_allowed = True

        # allow mutation (see frozen as well)
        allow_mutation = True

        # pass around models by ref
        copy_on_model_validation = "none"

        # no undeclared fields
        extra = "forbid"

        # not frozen
        # frozen = False

        # access by attr
        orm_mode = True

        # json_encoders = {}
        json_dumps = orjson_dumps
        json_loads = orjson_loads

    def __hash__(self):
        return hash(self.id)

    # disallow mutation
    def freeze(self):
        # TODO unfreeze?
        self._frozen = True

    def __setattr__(self, name, value):
        if self._frozen:
            raise TypeError(f'"{self.__class__.__name__}" is frozen and does not support item assignment')
        # attach transport if its a basemodel
        # TODO
        # if self._transport and isinstance(value, BaseModel):
        #     value.onTransport(self._transport)
        super().__setattr__(name, value)

    # overload copy to replace id
    def copy(self, clone=False, freeze=False, *args, **kwargs):
        copied = super().copy(*args, **kwargs)

        if not clone:
            # assign new unique id
            copied.id = str(uuid4())

            # overload last_updated
            copied.modified = utcnow()

            if freeze:
                # make readonly
                # NOTE: can only do if not cloning
                copied.freeze()

        return copied

    @classmethod
    def _walk_types(cls):
        for field in cls.__fields__.values():
            # TODO is this good enough?
            yield field.type_

    # transport layer
    _transport: "Transport" = PrivateAttr(default=None)

    def onTransport(self, transport: "Transport", checked: Optional[set] = None):
        """recursively attach transport"""
        # attach transport to self
        self._transport = transport

        to_check: deque[Any] = deque()
        checked = set([self] + list(checked or set()))
        to_check.extend(list(self.__dict__.values()))

        while to_check:
            # bfs
            value_to_check = to_check.popleft()

            # recurse if its a basemodel
            if isinstance(value_to_check, BaseModel):
                if value_to_check not in checked and value_to_check._transport is None:
                    value_to_check.onTransport(transport=transport, checked=checked)
                # add self to checked list
                checked.add(value_to_check)

            # try to extract if its iterable in some form
            # check if dict first to check key and value
            if isinstance(value_to_check, dict):
                to_check.extend(list(value_to_check.keys()))
                to_check.extend(list(value_to_check.values()))
            elif isinstance(value_to_check, (list, set, tuple)):
                to_check.extend((e for e in value_to_check))
            else:
                # TODO: give up?
                ...

    def onUpdate(self, other: "BaseModel", **kwargs) -> None:
        # TODO json diff OR
        # monkeypatch setattr/setitem recursively >:-)
        ...

    def update(self, model: "BaseModel", model_target: str = ""):
        from .update import Update

        self.send(Update(model=model, model_target=model_target))

    def notifyConnect(self, client_id: str) -> None:
        # TODO disallow?
        ...

    def notifyDisconnect(self, client_id: str) -> None:
        # TODO cleanup?
        ...

    def send(self, update: "Update") -> None:
        self._out_queue.put_nowait(update)

    async def sendAsync(self, update: "Update") -> None:
        await self._out_queue.put(update)

    def get(self) -> "Update":
        # TODO
        return self._transport.run(self._out_queue.get())

    async def getAsync(self) -> "Update":
        return await self._out_queue.get()

    def receive(self, update: "Update") -> None:
        # TODO
        # to be overloaded
        self.onUpdate(update.model)

    async def receiveAsync(self, update: "Update") -> None:
        return self.receive(update=update)


# class ListModel(PydanticBaseModel):
#     __root__: List[BaseModel]
#     class Config:
#         json_dumps = orjson_dumps
#         json_loads = orjson_loads
# class DictModel(PydanticBaseModel):
#     __root__: Dict[Any, BaseModel]
#     class Config:
#         json_dumps = orjson_dumps
#         json_loads = orjson_loads



import orjson


def orjson_dumps(v, *, default):
    # orjson.dumps returns bytes, to match standard json.dumps we need to decode
    return orjson.dumps(v, default=default).decode()


def orjson_loads(*args, **kwargs):
    return orjson.loads(*args, **kwargs)



from asyncio import gather
from typing import Type
from .connection import Connection
from .model import BaseModel
from .transport import Transport


class Server(Connection):
    _transport: Transport
    _transport_type: Type
    _model: BaseModel
    _client_id: str
    _shared: bool
    _readonly: bool

    def __init__(
        self,
        transport: Transport,
        model: BaseModel,
        shared: bool = True,
        readonly: bool = False,
        **kwargs,
    ):
        self._transport = transport
        self._transport_type = transport.type
        self._model = model
        self._client_id = ""
        self._shared = shared
        self._readonly = readonly

    async def onOpen(self):
        # wait for client connection
        self._client_id = await self.connect() or ""

        # Notify transport of connection
        self._client_id = await self._transport.onConnect(client_id=self._client_id)

        # Host model
        await self._transport.host(model=self._model, client_id=self._client_id, shared=self._shared, readonly=self._readonly)

        # send initial
        initial = self._transport.initial(client_id=self._client_id)

        # send to client
        await self.send(initial)

    async def onClose(self):
        # Notify transport of connection
        await self._transport.onDisconnect(model=self._model, client_id=self._client_id)
        # await self.disconnect()

    async def handle(self) -> None:
        try:
            # run the open routing
            await self.onOpen()

            # Now enter an infinite loop listening for updates from client/server forever
            await gather(self.sender(), self.receiver(), return_exceptions=True)
        except:
            raise
        finally:
            # run the on close routing
            await self.onClose()
